/* VOCODE - old-fashioned channel vocoder.

   Performs a filter-bank analysis of the rtinput file, and uses the time-
   varying energy measured in the filter bands to control a corresponding
   filter bank that processes an excitation signal. This signal comes either
   from a regular sound file or from a noise oscillator. The two filter banks
   have identical characteristics. There are two ways to configure the banks;
   the meanings of pfields 4-6 are different for the two ways.

   Open the analysis sound file with rtinput.

   To use a sound file as the excitation, call the excitationfile() Minc
   function, which takes the file name as its argument. (Call this before
   calling VOCODE.) If you don't specify an excitation file, then VOCODE
   uses noise instead. The meanings of pfields 9 and 10 change depending
   on whether you use a file or noise.

   p0  = output start time
   p1  = analysis input file start time
   p2  = duration
   p3  = amplitude multiplier (post-processing)

   Two ways to specify filter bank center frequencies:

    (1) Spread evenly above a given center frequency:
          p4 = number of filters (greater than 0)
          p5 = lowest filter center frequency (in Hz or oct.pc)
          p6 = center frequency spacing multiplier (greater than 1)
               (multiplies each cf by this to get next higher cf)

    (2) A list of center frequencies, given in function table 2 (Use gen 2.)
          p4 = 0 (must be zero: tells program to look for function table)
          p5 = transposition of function table, in oct.pc
          p6 = (ignored)
        Number of filters determined by length of function table.

   p7  = filter bandwidth proportion of center frequency (greater than 0)
   p8  = filter response time (seconds) [optional; default is 0.5]
         Determines how often changes in analysis file power are measured.

   Two ways to specify excitation function:

    (1) To use a sound file, first call excitationfile(), then:
          p9  = excitation file start time
          p10 = excitation file channel  [optional; default is 0]

    (2) To use noise, call excitationfile(""), or don't call it at all, then:
          p9  = seed for random number generator (any positive integer --
                   0 takes seed from system clock)
          p10 = specifies how often (in samples) to get new random values
                from the noise generator. Default is a new value every sample.

   p11 = analysis input file channel  [optional; default is 0]
   p12 = stereo spread (0 - 1)  [optional; default is 0.5]

   Assumes function table 1 is an amplitude curve for the note. (Try gen 18.)
   Or you can just call setline. If no setline or function table 1, uses a
   flat amplitude curve. This curve, combined with the amplitude multiplier,
   affect the signal after processing by the filter bank.

   NOTES:

     -  If using first method for specifying center frequencies...
           p6 = 2.0 will make a stack of octaves
           p6 = 1.5 will make a stack of perfect (Pythagorian) fifths
        Use this to get stacks of an equal tempered interval (in oct.pc):
           p6 = cpspch(interval) / cpspch(0.0)

     - If using second method for specifying center frequencies...
       Make function table 2, using gen 2. For example, to make cf's
       from a list of pitches in oct.pc notation:
          num_filters = 5
          makegen(2, 2, num_filters, 0)
          8.00, 8.07, 9.00, 9.07, 10.02
       You can transpose these by the number of semitones given in p5.
       Or, to specify cf's in Hz:
          makegen(2, 2, 12, 0)
          100, 200, 300, 400, 500, 600, 700, 800, 900, 1000, 1100, 1200
       Transposition (still specified as oct.pc) works here also, and it
       preserves harmonic structure.

     - If you want to use noise as excitation after having used a file,
       just say excitationfile("").

     - More about p10 (subsample interval) when using noise as excitation:
       When the interval is greater than 1, successive noise samples are
       connected using linear interpolation. This acts as a low-pass filter;
       the higher the interval, the lower the cutoff frequency. Above an
       interval of about 10 samples, the noise starts to sound gritty.


   John Gibson (jgg9c@virginia.edu), 7/19/99.
*/
#include <stdio.h>
#include <stdlib.h>
#include <ugens.h>
#include <mixerr.h>
#include <Instrument.h>
#include "VOCODE.h"
#include <rt.h>
#include <rtdefs.h>

extern "C" {
   #include "excitationfile.h"
}


VOCODE :: VOCODE() : Instrument()
{
   in = NULL;
   exciter_file = NULL;    // might not use below, so make sure it's null here
}


VOCODE :: ~VOCODE()
{
   delete [] in;
   delete exciter_file;
   for (int i = 0; i < numfilts; i++) {
      delete exciter_filt[i];
      delete analysis_filt[i];
      delete balancer[i];
   }
}


/* Note: Originally, I had a complicated scheme for handling more than one
   channel in either file, thinking that if both were stereo, the job would
   run faster than two separate mono calls to the instrument. No dice ... no
   significant difference, and the code is much simpler this way.   -JGG
*/

int VOCODE :: init(float p[], int n_args)
{
   int   j, exchans, balance_window;
   float outskip, analinskip, exciteinskip, dur;
   float lowcf, spacemult, bwpct, responsetime, cf[MAXFILTS];
   char  *exciter_filename;

   outskip = p[0];
   analinskip = p[1];
   dur = p[2];
   amp = p[3];
   numfilts = (int)p[4];
   lowcf = p[5];
   spacemult = p[6];
   bwpct = p[7];
   responsetime = n_args > 8 ? p[8] : 0.5;            /* default: .5 secs */
   exciteinskip = p[9];
   excitechan = (int)p[10];
   analchan = (int)p[11];                             /* default: 0 */
   spread = n_args > 12 ? p[12] : 0.5;                /* default: center */

   nsamps = rtsetoutput(outskip, dur, this);
   rtsetinput(analinskip, this);

   if (analchan >= inputchans)
      die("VOCODE", "You asked for channel %d of a %d-channel file.",
                                                       analchan, inputchans);
   exciter_filename = get_filename();
   if (exciter_filename) {
      exciter_file = new SoundIn(exciter_filename, exciteinskip);
      exchans = exciter_file->getChannels();
      excite_from_file = 1;
      if (excitechan >= exchans)
         die("VOCODE", "You asked for channel %d of a %d-channel "
                                    "excitation file.", excitechan, exchans);
   }
   else {
      int subsample = excitechan;       // pfield meanings change
      unsigned int seed = (unsigned int)exciteinskip;
      if (subsample < 1)
         subsample = 1;
      noise = new SubNoiseL(subsample, seed);
      excite_from_file = 0;
      advise("VOCODE", "No excitation file, so using noise instead.");
   }

   if (bwpct <= 0.0)
      die("VOCODE", "Bandwidth proportion must be greater than 0.");

   balance_window = (int)(responsetime * SR);
   if (balance_window < 2) {
      warn("VOCODE", "Response time too short ... changing to %.8f.", 2.0 / SR);
      /* Otherwise, can get ear-splitting output. */
      balance_window = 2;
   }

   /* <numfilts> pfield lets user specify filter center frequencies by
      interval, in the form of a frequency multiplier, or by function table.
   */
   if (numfilts > 0) {   /* by interval */
      if (numfilts > MAXFILTS)
         die("VOCODE", "Can only use %d filters.", MAXFILTS);
      if (spacemult <= 1.0)
         die("VOCODE",
                    "Center frequency spacing factor must be greater than 1.");
      if (lowcf < 15.0)                        /* interpreted as oct.pc */
         lowcf = cpspch(lowcf);

      advise("VOCODE", "Building center freqs above %g...", lowcf);
      for (j = 0; j < numfilts; j++)
         cf[j] = lowcf * (float)pow((double)spacemult, (double)j);
   }
   else {                /* by function table */
      float transp = lowcf;                    /* pfield meaning changes */
      float *freqtable = floc(2);
      if (freqtable == NULL)
         die("VOCODE", "You haven't made the center freq. function (table 2).");
      numfilts = fsize(2);
      if (numfilts > MAXFILTS)
         die("VOCODE", "Can only use %d filters.", MAXFILTS);
      advise("VOCODE", "Reading center freqs from function table...");
      for (j = 0; j < numfilts; j++) {
         float freq = freqtable[j];
         if (freq < 15.0) {                    /* interpreted as oct.pc */
            if (transp)
               cf[j] = cpsoct(octpch(freq) + octpch(transp));
            else
               cf[j] = cpspch(freq);
         }
         else {                                /* interpreted as Hz */
            if (transp)
               cf[j] = cpsoct(octcps(freq) + octpch(transp));
            else
               cf[j] = freq;
         }
      }
   }

   for (j = 0; j < numfilts; j++) {
      if (cf[j] > SR * 0.5) {
         warn("VOCODE", "A cf was above Nyquist. Correcting...");
         cf[j] = SR * 0.5;
      }
   }

   advise("VOCODE", "centerfreq  bandwidth");
   for (j = 0; j < numfilts; j++)
      advise(NULL, "         %10.4f %10.4f", cf[j], bwpct * cf[j]);

   for (j = 0; j < numfilts; j++) {
      analysis_filt[j] = new Butter();
      analysis_filt[j]->setBandPass(cf[j], bwpct * cf[j]);

      exciter_filt[j] = new Butter();
      exciter_filt[j]->setBandPass(cf[j], bwpct * cf[j]);

      balancer[j] = new Balance();
      balancer[j]->setWindowSize(balance_window);
      balancer[j]->setInitialGain(0.0);
   }

   amparray = floc(1);
   if (amparray) {
      int lenamp = fsize(1);
      tableset(dur, lenamp, amptabs);
   }
   else
      advise("VOCODE", "Setting phrase curve to all 1's.");

   skip = (int)(SR / (float)resetval);

   return nsamps;
}


int VOCODE :: run()
{
   int   i, j, branch, rsamps;
   float aamp, analsig, excitesig;
   float out[MAXBUS];

   if (in == NULL)              /* first time, so allocate it */
      in = new float [RTBUFSAMPS * inputchans];

   Instrument :: run();

   rsamps = chunksamps * inputchans;
   rtgetin(in, this, rsamps);

   aamp = amp;                  /* in case amparray == NULL */

   branch = 0;
   for (i = 0; i < rsamps; i += inputchans) {
      if (--branch < 0) {
         if (amparray)
            aamp = tablei(cursamp, amparray, amptabs) * amp;
         branch = skip;
      }
      analsig = in[i + analchan];
      if (excite_from_file) {
         float *sigs = exciter_file->tick();  /* returns one frame's worth */
         excitesig = sigs[excitechan];
      }
      else
         excitesig = noise->tick();

      out[0] = 0.0;
      for (j = 0; j < numfilts; j++) {
         float anal = analysis_filt[j]->tick(analsig);
         float exc = exciter_filt[j]->tick(excitesig);
         out[0] += balancer[j]->tick(exc, anal);
      }
      out[0] *= aamp;
      if (outputchans == 2) {
         out[1] = out[0] * (1.0 - spread);
         out[0] *= spread;
      }
      // FIXME: too bad if outputchans > 2

      rtaddout(out);
      cursamp++;
   }

   return i;
}


Instrument *makeVOCODE()
{
   VOCODE *inst;

   inst = new VOCODE();
   inst->set_bus_config("VOCODE");

   return inst;
}


void
rtprofile()
{
   RT_INTRO("VOCODE", makeVOCODE);
}


