/* transpan - modified version of Doug Scott's trans that does static
   panning from a mono input

   p0 = output start time
   p1 = input start time
   p2 = input duration (if negative, time to stop reading on input file)
   p3 = amplitude multiplier
   p4 = interval of transposition (8ve.pc)
   p5 = input channel number (from 0) [optional, default is 0]
   p6 = stereo position ("percent" to left, 0 - 1) [optional, default is .5]

   Assumes function table 1 is amplitude curve for the note. (Try gen 18.)
   Or you can just call setline. If no setline or function table 1, uses
   flat amplitude curve.

   The input file can be either stereo or mono. If mono, p6 (stereo pos.)
   is ignored.

   John Gibson <jgg9c@virginia.edu>, 2/8/00
*/
#include <ugens.h>
#include <sfheader.h>
#include <math.h>
#include <sndlibsupport.h> /* must come after math.h, or else M_PI redefined */
#include <sys/file.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>

#define INPUT  0
#define OUTPUT 1

extern float SR;
extern SFHEADER sfdesc[NFILES];
extern struct stat sfst[NFILES];


/* --------------------------------------------------------------- interp --- */
static double
interp(double y0, double y1, double y2, double t)
{
   double hy2, hy0, a, b, c;

   a = y0;
   hy0 = y0 / 2.0;
   hy2 = y2 / 2.0;
   b = (-3.0 * hy0) + (2.0 * y1) - hy2;
   c = hy0 - y1 + hy2;

   return (a + b * t + c * t * t);
}


/* ------------------------------------------------------------- transpan --- */
double
transpan(float p[], int n_args)
{
   int      j, z, getflag, incount, inchan, outchans;
   long     insamps, outsamps, outcount;
   float    outskip, inskip, indur, trans_pch, amp, aamp, pctleft;
   float    in[SF_MAXCHAN], out[2];          /* mono or stereo out only */
   float    *amparray, tabs[2];              /* for amp envelope */
   double   increment, counter, frac, voldsig, oldsig, newsig, dur, interval;

   if (n_args < 5)
      die("transpan", "Wrong number of args.");

   outskip = p[0];
   inskip = p[1];
   indur = p[2];
   amp = p[3];
   trans_pch = p[4];
   inchan = n_args > 5 ? (int)p[5] : 0;               /* default is chan 0 */
   pctleft = n_args > 6 ? p[6] : .5;

   if (inchan >= sfchans(&sfdesc[INPUT]))
      die("transpan", "You asked for channel %d of a %d-channel file.",
                                           inchan, sfchans(&sfdesc[INPUT]));

   outchans = sfchans(&sfdesc[OUTPUT]);
   if (outchans > 2)
      die("transpan", "Output file can't have more than 2 channels.");

   interval = octpch(trans_pch);           /* convert interval to lin octave */
   increment = cpsoct(10.0 + interval) / cpsoct(10.0);  /* the samp incr. */

   if (indur == 0.0) {
      indur = ( (float) (sfst[INPUT].st_size - getheadersize(&sfdesc[INPUT]))
              / (float) sfclass(&sfdesc[INPUT])
              / (float) sfchans(&sfdesc[INPUT])
              / sfsrate(&sfdesc[INPUT]) ) - inskip;
   }
   if (indur < 0.0)
      dur = (double) (-indur - inskip);                    /* timeend if < 0 */
   else
      dur = (double) indur;

   insamps = setnote(inskip, dur, INPUT);
   outsamps = setnote(outskip, (dur / increment), OUTPUT);

   amparray = floc(1);
   if (amparray) {
      int len = fsize(1);
      tableset(dur / increment, len, tabs);  /* set up table for amp */
   }
   else
      advise("transpan", "Setting phrase curve to all 1's.");

   z = SR / (float) resetval;                  /* control rate for amp curve */
   j = 0;

   /* the processing loop for the input signal */

   voldsig = 0.0;
   GETIN(in, INPUT);
   oldsig = (double) in[inchan];
   GETIN(in, INPUT);
   newsig = (double) in[inchan];

   aamp = amp;                                /* in case amparray == NULL */
   counter = 0.0;
   getflag = incount = 1;
   outcount = 0;

   while (outcount++ < outsamps) {
      while (getflag) {
         GETIN(in, INPUT);                /* get input samp; rotate old ones */
         voldsig = oldsig;
         oldsig = newsig;
         newsig = in[inchan];
         incount++;
         if ((counter - (float) incount) < 0.5)
            getflag = 0;
      }
      frac = counter - incount + 2.0;  /* The interp value */
      while (!j--) {
         if (amparray)
            aamp = tablei(outcount, amparray, tabs) * amp;
         j = z;
      }
      out[0] = interp(voldsig, oldsig, newsig, frac) * aamp;
      if (outchans == 2) {
         out[1] = out[0] * (1.0 - pctleft);
         out[0] *= pctleft;
      }
      ADDOUT(out, OUTPUT);

      counter += increment;                /* keeps track of interp pointer */
      if ((counter - (float) incount) >= -0.5)
         getflag = 1;                      /* ready for samp */
   }

   endnote(OUTPUT);

   return 0.0;
}


int
profile()
{
   UG_INTRO("transpan", transpan);
   return 0;
}

