/* RTcmix  - Copyright (C) 2000  The RTcmix Development Team
   See ``AUTHORS'' for a list of contributors. See ``LICENSE'' for
   the license to this software and for a DISCLAIMER OF ALL WARRANTIES.
*/

/* For v2.3, by John Gibson. Based on SGI code by Brad Garton and Doug Scott,
   and Linux code by Dave Topper.
   Revised to include OS X support by John Gibson, 11/01.
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <assert.h>
#include <audio_port.h>
#include <sndlibsupport.h>

#ifdef ALSA
   #define setup_device  setup_alsa_device
#endif /* ALSA */
#ifdef OSS
   #define setup_device  setup_oss_device
#endif /* OSS */

/* #define DEBUG */


/* -------------------------------------------------------- LINUX support --- */
#ifdef LINUX

/* ---------------------------------------------------------- open_device --- */
/* Open the specified audio device, and return its file descriptor.
   If error on open, print error msg and return -1.

   <num> is a number appended to the BASE_SND_DEVICE string.
         If -1, then no number appended.

   <access_type> are the flags passed to open (e.g., O_WRONLY).

   Examples (if BASE_SND_DEVICE is "/dev/dsp"):
      open_device(2, O_RDWR);      ... opens "/dev/dsp2" for read/write access.
      open_device(-1, O_RDONLY);   ... opens "/dev/dsp" for read-only access.
*/
static int
open_device(int num, int access_type)
{
   int   fd;
   char  buf[256];

   if (num == -1)
      strcpy(buf, BASE_SND_DEVICE);
   else
      sprintf(buf, "%s%d", BASE_SND_DEVICE, num);

   fd = open(buf, access_type, 0);
   if (fd == -1) {
      char  str[256];
      sprintf(str, "open_device: (%s)", buf);
      perror(str);
   }

   return fd;
}


#ifdef ALSA

/* ---------------------------------------------------- setup_alsa_device --- */
/* Set up the audio device parameters. Return 0 if ok, -1 if error.
*/
static int
setup_alsa_device(
      int   fd,             /* file descriptor of device just opened */
      int   nchans,         /* number of chans for this device */
      int   srate,          /* sampling rate */
      int   num_frags,      /* number of buffer fragments */
      int   duplex,         /* 1 if you want a full duplex device */
      int   *buffer_size)   /* desired buffer size; actual size returned */
{
   fprintf(stderr, "ALSA driver not supported yet.\n");

   return 0;
}

#endif /* ALSA */


#ifdef OSS

/* --------------------------------------------------- get_card_max_chans --- */
int
get_card_max_chans(int fd)
{
   int card_chans;
#ifdef NEW_OSS

#if 0
   if (ioctl(fd, SOUND_PCM_READ_CHANNELS, &card_chans) == -1) {
      perror("SOUND_PCM_READ_CHANNELS");
      return -1;
   }
#else
//FIXME: need to ask OSS for clarification...

/* How do we get from OSS the max number of channels a card supports?
   Hannu says:
      "You can try to set the number of channels to the maximum number
      of channels your application supports. The ioctl() call will
      return back the maximum number of channels the device supports.
      Minimum number of channels can be found out by requesting 1
      channels. If the device doesn't support mono then the call will
      set the parameter to 2. After that you can set the number of
      channels to any value between the minimum and maximum."

   However, with OSS 3.9.1i and AudioPCI card, the ioctl passes
   back 1 when <card_chans> is > 2 (and returns 0).
   (Haven't tested with RME.)
*/
	card_chans = 8;
	if (ioctl(fd, SOUND_PCM_WRITE_CHANNELS, &card_chans) == -1) {
	   perror("SOUND_PCM_WRITE_CHANNELS");
	   return -1;
	}
	/* To address above issue, fall through to old code if card_chans == 1
	   and ioctl() did not fail */
	else if (card_chans == 1)	
#endif /* !0 */

#endif /* !NEW_OSS */
    {
	/* Try to set device to stereo mode the old way. */
	int stereo = 1;
	if (ioctl(fd, SNDCTL_DSP_STEREO, &stereo) == -1)
	   card_chans = 1;     /* assume it's mono only if it fails, but who cares then */
	else
	   card_chans = 2;
    }
	return card_chans;
}


/* ----------------------------------------------------- setup_oss_device --- */
/* Set up the audio device parameters. Return 0 if ok, -1 if error.
*/
static int
setup_oss_device(
      int   fd,             /* file descriptor of device just opened */
      int   nchans,         /* number of chans for this device */
      int   srate,          /* sampling rate */
      int   num_frags,      /* number of buffer fragments */
      int   duplex,         /* 1 if you want a full duplex device */
      int   *buffer_size)   /* desired buffer size; actual size returned */
{
   int      format, arg, stereo, tmp, frag_size, card_chans, rate;
   double   bufexponent;

   assert(fd > 0);

   /* Set buffering parameters (number of fragments and fragment size).
      OSS requires this way of specifying things.
   */
   arg = num_frags << 16;
   bufexponent = log((double) *buffer_size) / log(2.0);
   arg += (int) bufexponent;
#ifdef DEBUG
   printf("arg: %#10.8x\n", arg);
#endif
   if (ioctl(fd, SNDCTL_DSP_SETFRAGMENT, &arg) == -1) {
      perror("SNDCTL_DSP_SETFRAGMENT");
      return -1;
   }

   /* NOTE: Must set full duplex before any of the other stuff below. */
   /*   Checks if card supports full duplex (SNDCTL_DSP_GETCAPS) */
   if (duplex) {
   	  int caps = 0;
	  if (ioctl(fd, SNDCTL_DSP_GETCAPS, &caps) == -1) {
         perror("SNDCTL_DSP_GETCAPS");
         return -1;
      }
	  else if (!(caps && DSP_CAP_DUPLEX)) {
	  	 fprintf(stderr, "Audio device does not support full duplex.\n");
		 return -1;
	  }
   	  tmp = duplex;
      if (ioctl(fd, SNDCTL_DSP_SETDUPLEX, &tmp) == -1) {
         perror("SNDCTL_DSP_SETDUPLEX");
         return -1;
      }
	  else if (tmp != duplex) {
	  	 fprintf(stderr, "Audio device does not support full duplex.\n");
		 return -1;
	  }
   }

   /* Set sample format */
#if MUS_LITTLE_ENDIAN
   format = AFMT_S16_LE;
#else
   format = AFMT_S16_BE;
#endif
   if (ioctl(fd, SNDCTL_DSP_SETFMT, &format) == -1) {
      perror("SNDCTL_DSP_SETFMT");
      return -1;
   }

#ifdef NEW_OSS
 #ifdef USE_CARD_CHANS
   card_chans = get_card_max_chans(fd);
   if (card_chans < nchans) {
      fprintf(stderr, "Audio device does not support %d chans.\n", nchans);
	  return -1;
   }
 #endif
   // Now make the request to set channel count.
   card_chans = nchans;
   if (ioctl(fd, SNDCTL_DSP_CHANNELS, &card_chans) == -1) {
      perror("SNDCTL_DSP_CHANNELS");
      return -1;
   }
 #ifndef USE_CARD_CHANS
   // In this case we need to recheck the returned value.
   else if (card_chans != nchans) {
      fprintf(stderr, "Audio device does not support %d chans.\n", nchans);
	  return -1;
   }
 #endif
#else /* !NEW_OSS */
   /* Set stereo / mono (1/0) */
   if (nchans != 1 && nchans != 2) {
      fprintf(stderr, "Audio device does not support %d chans.\n", nchans);
	  return -1;
   }
   stereo = nchans - 1;
   if (ioctl(fd, SNDCTL_DSP_STEREO, &stereo) == -1) {
      perror("SNDCTL_DSP_STEREO");
      return -1;
   }
#endif /* !NEW_OSS */

   /* Set sample rate */
   rate = srate;
   if (ioctl(fd, SNDCTL_DSP_SPEED, &rate) == -1) {
      perror("SNDCTL_DSP_SPEED");
      return -1;
   }
   else if (rate != srate) {
      fprintf(stderr, "Warning:  Audio device rate %d != requested rate %d\n", 
	  		  rate, srate);
   }

   /* Get size of audio buffer. Note that driver computes this optimally. */
   if (ioctl(fd, SNDCTL_DSP_GETBLKSIZE, &frag_size) == -1) {
      perror("SNDCTL_DSP_GETBLKSIZE");
      return -1;
   }
#ifdef DEBUG
   printf("Buffer size: %d (requested), %d (actual)\n",
                                                    *buffer_size, frag_size);
#endif

   return 0;
}

#endif /* OSS */


/* ----------------------------------------------------------- open_ports --- */
int
open_ports(
      int      in_nchans,         /* 0 for no input */
      int      max_inports,       /* size of in_ports[] */
      int      in_ports[],        /* array of file descriptors */
      int      out_nchans,        /* 0 for no output */
      int      max_outports,      /* size of out_ports[] */
      int      out_ports[],       /* array of file descriptors */
      int      verbose,           /* print buffer sizes, etc. */
      float    srate,
      int      nfrags,            /* number of fragments to request */
      int      *nframes)          /* number of frames per buffer to request; */
                                  /* on return, contains actual frames */
{
   int   n, status, datum_size, nchans, in_buf_bytes, out_buf_bytes;

   datum_size = AUDIO_DATUM_SIZE;

#ifdef MONO_DEVICES

   in_buf_bytes = out_buf_bytes = *nframes * datum_size;

   if (in_nchans > 0) {
      assert(in_nchans <= max_inports);

      for (n = 0; n < in_nchans; n++) {
         in_ports[n] = open_device(n + IN_DEV_OFFSET, O_RDONLY);
         if (in_ports[n] == -1)
            return -1;

         status = setup_device(in_ports[n], 1, (int) srate, nfrags, 0,
                                                              &in_buf_bytes);
         if (status == -1)
            return -1;
      }
   }

   if (out_nchans > 0) {
      assert(out_nchans <= max_outports);

      for (n = 0; n < out_nchans; n++) {
         out_ports[n] = open_device(n + OUT_DEV_OFFSET, O_WRONLY);
         if (out_ports[n] == -1)
            return -1;

         status = setup_device(out_ports[n], 1, (int) srate, nfrags, 0,
                                                              &out_buf_bytes);
         if (status == -1)
            return -1;
      }
   }

#else /* !MONO_DEVICES */

   in_buf_bytes = *nframes * in_nchans * datum_size;
   out_buf_bytes = *nframes * out_nchans * datum_size;

 #ifdef SHARED_AUDIO_DEV

   assert(max_outports >= 1);

   if (out_nchans > 0 && in_nchans > 0) {
      assert(out_nchans == in_nchans);

      out_ports[0] = open_device(-1, O_RDWR);
      if (out_ports[0] == -1)
         return -1;

      status = setup_device(out_ports[0], out_nchans, (int) srate,
                                                   nfrags, 1, &out_buf_bytes);
      if (status == -1)
         return -1;
      in_ports[0] = out_ports[0];
      in_buf_bytes = out_buf_bytes;    /* setup_device can change this */
   }
   else if (out_nchans > 0) {
      out_ports[0] = open_device(-1, O_WRONLY);
      if (out_ports[0] == -1)
         return -1;

      status = setup_device(out_ports[0], out_nchans, (int) srate,
                                                   nfrags, 0, &out_buf_bytes);
      if (status == -1)
         return -1;
   }
   // else: no input-only mode

 #else /* !SHARED_AUDIO_DEV */

   if (in_nchans > 0) {
      assert(max_inports >= 1);

      in_ports[0] = open_device(-1, O_RDONLY);
      if (in_ports[0] == -1)
         return -1;

      status = setup_device(in_ports[0], in_nchans, (int) srate,
                                                   nfrags, 0, &in_buf_bytes);
      if (status == -1)
         return -1;
   }

   if (out_nchans > 0) {
      assert(max_outports >= 1);

      out_ports[0] = open_device(-1, O_WRONLY);
      if (out_ports[0] == -1)
         return -1;

      status = setup_device(out_ports[0], out_nchans, (int) srate,
                                                   nfrags, 0, &out_buf_bytes);
      if (status == -1)
         return -1;
   }

 #endif /* !SHARED_AUDIO_DEV */

#endif /* !MONO_DEVICES */

   /* OSS can change the requested buffer size.  -JGG */
   nchans = out_nchans > 0 ? out_nchans : in_nchans;
   *nframes = (out_buf_bytes / datum_size) / nchans;

   if (verbose) {
      if (out_nchans > 0)
         printf("Audio output buffer set to %d frames (%d bytes).\n",
                                                   *nframes, out_buf_bytes);
      if (in_nchans > 0)
         printf("Audio input buffer set to %d bytes.\n", in_buf_bytes);
   }

   return 0;
}

#endif /* LINUX */


/* ------------------------------------------------------- MACOSX support --- */
#ifdef MACOSX
#include <buffers.h>
#include <pthread.h>
#include <mach/mach.h>
#include <mach/policy.h>

/* For signalling between main thread and IOProc, which runs as a higher
   priority (IOThread) thread. 
*/
static pthread_cond_t audio_output_condition;
static int buffer_ready = 0;
static float *intermediate_in_buffer, *intermediate_out_buffer;
extern BufPtr audioin_buffer[];  /* avoid pulling in all of globals.h */
extern BufPtr out_buffer[];
extern int play_audio;

typedef struct {
   int   nfrags;
   int   nframes;
   int   in_nchans;
   int   out_nchans;
   int   indev_nchans;
   int   outdev_nchans;
} DevInfo;
static DevInfo dev_info;

/* ---------------------------------------------------------- audioIOProc --- */
#define IN_GAIN_FACTOR  32768.0
#define OUT_GAIN_FACTOR (1.0 / 32768.0)

static OSStatus
audioIOProc(
      AudioDeviceID           inDevice,
      const AudioTimeStamp    *inNow,
      const AudioBufferList   *inInputData,
      const AudioTimeStamp    *inInputTime,
      AudioBufferList         *outOutputData,
      const AudioTimeStamp    *inOutputTime,
      void                    *devinfo)
{
   int      nsamps;
   float    *in, *out, *tmp;
   DevInfo  *info = (DevInfo *) devinfo;
   static int cur_fragment = 0;
#ifdef DEBUG
   static Float64 lasttime=-1, thistime, timediff;
#endif

   if (!play_audio || !buffer_ready)
      return kAudioHardwareNoError;

#ifdef DEBUG
   thistime = inOutputTime->mSampleTime;
   if (lasttime==-1)
      timediff=0;
   else {
      timediff = thistime - lasttime;
      if (timediff!=info->nframes)
         printf("******timediff != buffer size\n");
   }
   lasttime = thistime;
   printf("ioproc: now=%g, thistime=%g, timediff=%g\n",
               inNow->mSampleTime, thistime, timediff);
   printf("        inchans=%ld, outchans=%ld, "
                              "inbytes=%ld, outbytes=%ld, curfrag=%d\n",
               inInputData->mBuffers[0].mNumberChannels,
               outOutputData->mBuffers[0].mNumberChannels,
               inInputData->mBuffers[0].mDataByteSize,
               outOutputData->mBuffers[0].mDataByteSize,
               cur_fragment);
   fflush(stdout);
#endif

   nsamps = info->nframes * info->outdev_nchans;
   tmp = intermediate_out_buffer + (cur_fragment * nsamps);
//FIXME: What if device has multiple, non-interleaved buffers?
   out = outOutputData->mBuffers[0].mData;
   while (nsamps--)
      *out++ = *tmp++;

   if (info->in_nchans > 0) {
      nsamps = info->nframes * info->indev_nchans;
      tmp = intermediate_in_buffer + (cur_fragment * nsamps);
//FIXME: What if device has multiple, non-interleaved buffers?
      in = inInputData->mBuffers[0].mData;
      while (nsamps--)
         *tmp++ = *in++;
   }

   cur_fragment++;
   if (cur_fragment == info->nfrags) {
      /* unblock main thread waiting on this condition variable */
      pthread_cond_signal(&audio_output_condition);
      cur_fragment = 0;
   }

   return kAudioHardwareNoError;
}


/* ---------------------------------------------------- macosx_audio_read --- */
void
macosx_audio_read()
{
   float *in;
   int   i, j, nsamps;
   static int cur_fragment = 0;

   if (dev_info.in_nchans == 0)
      return;

#ifdef DEBUG
   printf(" read: curfrag=%d\n", cur_fragment);
   fflush(stdout);
#endif

   nsamps = dev_info.nframes * dev_info.indev_nchans;
   in = intermediate_in_buffer + (cur_fragment * nsamps);

   /* Copy from buffer fragment, with appropriate channel mapping. */
   if (dev_info.indev_nchans == dev_info.in_nchans) {
      for (i = 0; i < dev_info.nframes; i++)
         for (j = 0; j < dev_info.in_nchans; j++)
            audioin_buffer[j][i] = *in++ * IN_GAIN_FACTOR;
   }
   else if (dev_info.indev_nchans > dev_info.in_nchans) {
      for (i = 0; i < dev_info.nframes; i++) {
         for (j = 0; j < dev_info.in_nchans; j++)
            audioin_buffer[j][i] = *in++ * IN_GAIN_FACTOR;
         for ( ; j < dev_info.indev_nchans; j++)   /* skip remaining chans */
            in++;
      }
   }
   else if (dev_info.indev_nchans < dev_info.in_nchans) {
      for (i = 0; i < dev_info.nframes; i++) {
         for (j = 0; j < dev_info.in_nchans; j++)
            audioin_buffer[j][i] = *in++ * IN_GAIN_FACTOR;
         for ( ; j < dev_info.in_nchans; j++)      /* zero remaining chans */
            audioin_buffer[j][i] = 0.0;
      }
   }

   cur_fragment++;
   if (cur_fragment == dev_info.nfrags)
      cur_fragment = 0;
}


/* --------------------------------------------------- macosx_audio_write --- */
void
macosx_audio_write()
{
   float *out;
   int   i, j, nsamps;
   static int cur_fragment = 0;
   static pthread_mutex_t audio_output_mutex = PTHREAD_MUTEX_INITIALIZER;

#ifdef DEBUG
   printf(" write: curfrag=%d\n", cur_fragment);
   fflush(stdout);
#endif

   nsamps = dev_info.nframes * dev_info.outdev_nchans;
   out = intermediate_out_buffer + (cur_fragment * nsamps);

   /* Copy into buffer fragment, with appropriate channel mapping. */
   if (dev_info.out_nchans == dev_info.outdev_nchans) {
      for (i = 0; i < dev_info.nframes; i++)
         for (j = 0; j < dev_info.out_nchans; j++)
            *out++ = out_buffer[j][i] * OUT_GAIN_FACTOR;
   }
   else if (dev_info.out_nchans == 1) {
      for (i = 0; i < dev_info.nframes; i++) {
         float samp = out_buffer[0][i] * OUT_GAIN_FACTOR;
         for (j = 0; j < dev_info.outdev_nchans; j++)
            *out++ = samp;
      }
   }
   else {
      for (i = 0; i < dev_info.nframes; i++) {
         for (j = 0; j < dev_info.out_nchans; j++)
            *out++ = out_buffer[j][i] * OUT_GAIN_FACTOR;
         for ( ; j < dev_info.outdev_nchans; j++) /* zero remaining dev chans */
            *out++ = 0.0;
      }
   }

   buffer_ready = 1;
   cur_fragment++;
   if (cur_fragment == dev_info.nfrags) {
      /* init wait condition */
      pthread_cond_init(&audio_output_condition, NULL);
      /* acquire lock */
      pthread_mutex_lock(&audio_output_mutex);
      /* wait until IOProc signals us to continue */
      pthread_cond_wait(&audio_output_condition, &audio_output_mutex);
      /* relinquish lock */
      pthread_mutex_unlock(&audio_output_mutex);
      cur_fragment = 0;
   }
}


/* ------------------------------------------ macosx_cmixplay_audio_write --- */
void
macosx_cmixplay_audio_write(short *buf)
{
   float *out;
   short *in;
   int   i, j, nsamps;
   static int cur_fragment = 0;
   static pthread_mutex_t audio_output_mutex = PTHREAD_MUTEX_INITIALIZER;

#ifdef DEBUG
   printf(" write: curfrag=%d\n", cur_fragment);
   fflush(stdout);
#endif

   nsamps = dev_info.nframes * dev_info.outdev_nchans;
   out = intermediate_out_buffer + (cur_fragment * nsamps);
   in = buf;

   /* Copy into buffer fragment, with appropriate channel mapping. */
   if (dev_info.outdev_nchans == dev_info.out_nchans) {
      for (i = 0; i < nsamps; i++)
         *out++ = *in++ * OUT_GAIN_FACTOR;
   }
   else if (dev_info.out_nchans == 1) {
      for (i = 0; i < dev_info.nframes; i++) {
         float samp = *in++ * OUT_GAIN_FACTOR;
         for (j = 0; j < dev_info.outdev_nchans; j++)
            *out++ = samp;
      }
   }
#if 0
   else {
      for (i = 0; i < dev_info.nframes; i++) {
         for (j = 0; j < dev_info.out_nchans; j++)
            *out++ = out_buffer[j][i] * OUT_GAIN_FACTOR;
         for ( ; j < dev_info.outdev_nchans; j++) /* zero remaining dev chans */
            *out++ = 0.0;
      }
   }
#endif

   buffer_ready = 1;
   cur_fragment++;
   if (cur_fragment == dev_info.nfrags) {
      /* init wait condition */
      pthread_cond_init(&audio_output_condition, NULL);
      /* acquire lock */
      pthread_mutex_lock(&audio_output_mutex);
      /* wait until IOProc signals us to continue */
      pthread_cond_wait(&audio_output_condition, &audio_output_mutex);
      /* relinquish lock */
      pthread_mutex_unlock(&audio_output_mutex);
      cur_fragment = 0;
   }
}


/* ---------------------------------------------------- open_macosx_ports --- */
#define ISINPUT 1
#define ISOUTPUT 0

int
open_macosx_ports(
      int            in_nchans,   /* 0 for no input */
      int            out_nchans,  /* 0 for no output */
      AudioDeviceID  *iport,      /* input device ID */
      AudioDeviceID  *oport,      /* output device ID */
      int            verbose,     /* print buffer sizes, etc. */
      float          srate,
      int            nfrags,      /* number of buffer fragments */
      int            *nframes)    /* number of frames per buffer to request; */
                                  /* on return, contains actual frames */
{
   int         indev_nchans, outdev_nchans, nsamps;
   UInt32      count, in_buffer_size, out_buffer_size, desired_buffer_size;
   float       host_clock_rate;
   uint32_t    period;
   thread_time_constraint_policy_data_t thread_policy;
   char        outdev_name[256];
   OSStatus    err = kAudioHardwareNoError;
   Boolean     writeable, running, alive;
   AudioDeviceID  indevice, outdevice;
   AudioStreamBasicDescription devFormat;

   if (!play_audio)
      return 0;

   indevice = outdevice = kAudioDeviceUnknown;
   if (iport)
      *iport = kAudioDeviceUnknown;
   if (oport)
      *oport = kAudioDeviceUnknown;

   /* Get default audio output device. */
   count = sizeof(outdevice);
   err = AudioHardwareGetProperty(kAudioHardwarePropertyDefaultOutputDevice,
               &count, (void *) &outdevice);
   if (err != kAudioHardwareNoError || outdevice == kAudioDeviceUnknown) {
      fprintf(stderr, "Can't find default audio output device (error %ld).\n",
                                                                         err);
      return -1;
   }

#ifdef NOMORE
   /* Get audio device name. */
   count = 256;
   err = AudioDeviceGetProperty(outdevice, 0, ISOUTPUT,
                     kAudioDevicePropertyDeviceName, &count, outdev_name);
   assert(err == kAudioHardwareNoError);
   printf("Audio output device: \"%s\"\n",  outdev_name);
#endif

#ifdef NOT_HERE
// FIXME: Do this from a notification callback?
   /* Make sure device is alive (connected). */
   count = sizeof(UInt32);
   err = AudioDeviceGetProperty(outdevice, 0, ISOUTPUT,
                     kAudioDevicePropertyDeviceIsAlive, &count, &alive);
   assert(err == kAudioHardwareNoError);
   if (!alive) {
      fprintf(stderr, "Audio output device not connected.\n");
      return -1;
   }

   /* Make sure device is running.  (What does this mean, exactly?) */
   count = sizeof(UInt32);
   err = AudioDeviceGetProperty(outdevice, 0, ISOUTPUT,
                     kAudioDevicePropertyDeviceIsRunning, &count, &running);
   assert(err == kAudioHardwareNoError);
   if (!running) {
      fprintf(stderr, "Audio output device not running.\n");
      return -1;
   }
#endif

   /* Get format information for output device, and check that it will
      work for our purposes.
   */
   count = sizeof(devFormat);
   err = AudioDeviceGetProperty(outdevice, 0, ISOUTPUT,
                     kAudioDevicePropertyStreamFormat, &count, &devFormat);
   if (err != kAudioHardwareNoError) {
      fprintf(stderr, "Can't get audio output device format (error %ld)\n",
                                                                         err);
      return -1;
   }
   assert(devFormat.mFormatID == kAudioFormatLinearPCM);
   assert(devFormat.mFormatFlags & kLinearPCMFormatFlagIsFloat);

   if (devFormat.mSampleRate != srate) {
      fprintf(stderr, "The sampling rate used by the audio output device "
                      "(%.2f kHz) doesn't match the requested rate "
                      "(%.2f kHz).\n",
                      devFormat.mSampleRate * .001, srate * .001);
      /* And note that we can't change the dev rate (according to penrose). */
      return -1;
   }

   outdev_nchans = devFormat.mChannelsPerFrame;
   if (out_nchans > outdev_nchans) {
      fprintf(stderr, "The audio output device has %d channels, but RTcmix "
                      "wants %d channels.\n", outdev_nchans, out_nchans);
      return -1;
   }

   /* Try to set buffer size for this device in accord with RTBUFSAMPS. */
   err = AudioDeviceGetPropertyInfo(outdevice, 0, ISOUTPUT,
            kAudioDevicePropertyBufferSize, &count, &writeable);
   assert(err == kAudioHardwareNoError);
   desired_buffer_size = *nframes * outdev_nchans * sizeof(float);
   if (writeable) {
      err = AudioDeviceSetProperty(outdevice, NULL, 0, ISOUTPUT,
                        kAudioDevicePropertyBufferSize,
                        count, (void *)&desired_buffer_size);
      /* NB: even if err != kAudioHardwareNoError, we continue and use
         the default buffer size.  */
   }

   /* Get the actual buffer size.  (The device may not want to change.) */
   err = AudioDeviceGetProperty(outdevice, 0, ISOUTPUT,
                     kAudioDevicePropertyBufferSize, &count, &out_buffer_size);
   if (err != kAudioHardwareNoError) {
      fprintf(stderr, "Can't get audio output device buffer size (error %ld)\n",
                                                                         err);
      return -1;
   }
   if (out_buffer_size != desired_buffer_size) {
		int desired_frames = desired_buffer_size / outdev_nchans / sizeof(float);
      *nframes = out_buffer_size / outdev_nchans / sizeof(float);
      printf("\nWARNING: Device using buffer size of %d frames (%lu bytes),\n"
             "instead of your requested buffer size of %d frames (%lu bytes)."
             "\n\n",
             *nframes, out_buffer_size, desired_frames, desired_buffer_size);
   }

//FIXME: other sanity tests?

   /* Open input device, if necessary. */
   if (in_nchans > 0) {

      /* Get default audio input device. */
      count = sizeof(indevice);
      err = AudioHardwareGetProperty(kAudioHardwarePropertyDefaultInputDevice,
                  &count, (void *) &indevice);
      if (err != kAudioHardwareNoError || indevice == kAudioDeviceUnknown) {
         fprintf(stderr, "Can't find default audio input device (error %ld).\n",
                                                                         err);
         return -1;
      }

      /* Get format information for input device, and check it. */
      count = sizeof(devFormat);
      err = AudioDeviceGetProperty(indevice, 0, ISINPUT,
                        kAudioDevicePropertyStreamFormat, &count, &devFormat);
      if (err != kAudioHardwareNoError) {
         fprintf(stderr, "Can't get audio input device format (error %ld).\n",
                                                                         err);
         return -1;
      }
      assert(devFormat.mFormatID == kAudioFormatLinearPCM);
      assert(devFormat.mFormatFlags & kLinearPCMFormatFlagIsFloat);

      if (devFormat.mSampleRate != srate) {
         fprintf(stderr, "The sampling rate used by the audio input device "
                         "(%.2f kHz) doesn't match the requested rate "
                         "(%.2f kHz).\n",
                         devFormat.mSampleRate * .001, srate * .001);
         return -1;
      }

      indev_nchans = devFormat.mChannelsPerFrame;
      if (in_nchans > indev_nchans) {
         fprintf(stderr, "The audio input device has %d channels, but RTcmix "
                         "wants %d channels.\n", indev_nchans, in_nchans);
         return -1;
      }

      /* Try to set buffer size for this device to match *inframes. */
      err = AudioDeviceGetPropertyInfo(indevice, 0, ISINPUT,
               kAudioDevicePropertyBufferSize, &count, &writeable);
      assert(err == kAudioHardwareNoError);
      desired_buffer_size = *nframes * indev_nchans * sizeof(float);
      if (writeable) {
         err = AudioDeviceSetProperty(indevice, NULL, 0, ISINPUT,
                           kAudioDevicePropertyBufferSize,
                           count, (void *)&desired_buffer_size);
         /* ??NB: even if err != kAudioHardwareNoError, we continue and use
            the default buffer size.  */
      }

      /* Get the actual buffer size.  (The device may not want to change.) */
      err = AudioDeviceGetProperty(indevice, 0, ISINPUT,
                     kAudioDevicePropertyBufferSize, &count, &in_buffer_size);
      if (err != kAudioHardwareNoError) {
         fprintf(stderr,
               "Can't get audio input device buffer size (error %ld).\n", err);
         return -1;
      }
      if (in_buffer_size != out_buffer_size) {
         fprintf(stderr, "Input device cannot be set to use the same number "
                         "of audio frames as output device.");
         return -1;
      }

      /* For unknown reasons, I get gapping with more than one fragment when
         audio input is on.  (Could have to do with synchronization of the
         various local cur_fragment variables in this file.)   -JGG
      */
      nfrags = 1;
   }
   else
      indev_nchans = 0;

   dev_info.nfrags = nfrags;
   dev_info.nframes = *nframes;
   dev_info.in_nchans = in_nchans;
   dev_info.out_nchans = out_nchans;
   dev_info.indev_nchans = indev_nchans;
   dev_info.outdev_nchans = outdev_nchans;

   /* Adjust scheduling policy for our thread.  Using "time constraint"
      scheduling prevents GUI manipulations from causing dropouts. */
   host_clock_rate = AudioGetHostClockFrequency();
   period = (uint32_t)( ((out_buffer_size / outdev_nchans)
                                 / devFormat.mSampleRate) * host_clock_rate );
   thread_policy.period = period;
   thread_policy.computation = AudioGetHostClockMinimumTimeDelta() * 2000;
   thread_policy.constraint = thread_policy.computation * 2;
   thread_policy.preemptible = 1;
   thread_policy_set(pthread_mach_thread_np(pthread_self()),
                     THREAD_TIME_CONSTRAINT_POLICY,
                     (thread_policy_t) &thread_policy,
                     THREAD_TIME_CONSTRAINT_POLICY_COUNT);

   /* Register our callback function with the HAL. */
   err = AudioDeviceAddIOProc(outdevice, audioIOProc, (void *)&dev_info);
   assert(err == kAudioHardwareNoError);

   /* Start audio input and output devices. */
   err = AudioDeviceStart(outdevice, audioIOProc);
   assert(err == kAudioHardwareNoError);
   if (in_nchans > 0) {
      err = AudioDeviceStart(indevice, audioIOProc);
      assert(err == kAudioHardwareNoError);
   }

   /* Allocate fragmented intermediate buffers. */
   nsamps = *nframes * outdev_nchans * nfrags;
   intermediate_out_buffer = (float *)calloc(nsamps, sizeof(float));
   assert(intermediate_out_buffer != NULL);
   if (in_nchans > 0) {
      nsamps = *nframes * indev_nchans * nfrags;
      intermediate_in_buffer = (float *)calloc(nsamps, sizeof(float));
      assert(intermediate_in_buffer != NULL);
   }

   if (verbose)
      printf("Audio output buffer set to %d frames (%ld bytes).\n",
                                                *nframes, out_buffer_size);

   if (iport)
      *iport = indevice;
   if (oport)
      *oport = outdevice;

   return 0;
}

#endif /* MACOSX */


/* ---------------------------------------------------------- SGI support --- */
#ifdef SGI

/* ------------------------------------------------------- open_sgi_ports --- */
/* Configure and open the output audio port specified by <out_port_str>,
   passing back the opened port as <oport>.
   We will write floats directly to the device, letting it know that they
   will be within short integer range so we do not need to limit.  This
   reduces the CPU overhead of the alWriteSamps call.
*/
// FIXME: We don't yet support this float-writing in the
//        Linux-derived SGI port.   -JGG
int
open_sgi_ports(
      char     *out_port_str,
      int      out_nchans,
      ALport   *oport,
      int      verbose,           /* (currently unused) */
      float    srate,
      int      *nframes)          /* number of frames per buffer to request; */
                                  /* on return, contains actual frames */
{
   long     pvbuf[2], buflen;
   ALconfig out_port_config;

   out_port_config = alNewConfig();

#ifdef AL_2_0
   /* Set output port if we specified one. */
   if (out_port_str != NULL) {
      int device = alGetResourceByName(AL_SYSTEM, out_port_str, AL_DEVICE_TYPE);
      if (device > 0)
         alSetDevice(out_port_config, device);
      else {
         fprintf(stderr, "WARNING: The audio port you specified: '%s', "
                         "is invalid\nUsing default port.\n", out_port_str);
         alSetDevice(out_port_config, AL_DEFAULT_DEVICE);
      }
   }
   alSetLimiting(out_port_config, 0);
#endif

#ifdef SEND_FLOATS_TO_DAC
   alSetSampFmt(out_port_config, AL_SAMPFMT_FLOAT);
   alSetFloatMax(out_port_config, 32768.0);
#else
   alSetSampFmt(out_port_config, AL_SAMPFMT_TWOSCOMP);
   alSetWidth(out_port_config, AL_SAMPLE_16);
#endif

   alSetChannels(out_port_config, out_nchans);

   if (alSetQueueSize(out_port_config, *nframes * 4) == -1) {
      fprintf(stderr, "Could not configure the output audio port queue "
                      "size to %d\n", *nframes * 4);
      alFreeConfig(out_port_config);
      return -1;
   }
   pvbuf[0] = AL_OUTPUT_RATE;
   pvbuf[1] = (long) srate;
   buflen = 2;
   ALsetparams(AL_DEFAULT_DEVICE, pvbuf, buflen);

// FIXME: We can't handle buffer size changing yet.   -JGG
#ifdef NOT_YET
   /* open up thet thar audio port! (if it has not been opened).
      otherwise, attempt to reconfigure the existing port.  If we have
      changed buffer size, we need to close and reopen.
   */
   if (*oport != NULL && newBufSamps != *nframes) {
      if (alClosePort(*oport) < 0) {
         fprintf(stderr,
                 "Could not close output audio port for reconfigure:  %s\n",
                                                alGetErrorString(oserror()));
         return -1;
      }
      *oport = NULL;
      *nframes = newBufSamps;
   }
#endif /* NOT_YET */
   if (*oport == NULL)
      *oport = alOpenPort("rtcmixout", "w", out_port_config);
   else if (alSetConfig(*oport, out_port_config) < 0) {
      fprintf(stderr, "Could not reconfigure output audio port:  %s\n",
                                                alGetErrorString(oserror()));
   }

   alFreeConfig(out_port_config);

   if (*oport == NULL) {
      fprintf(stderr, "Could not open output audio port:  %s\n",
                                                alGetErrorString(oserror()));
      exit(1);
   }

   return 0;
}

#endif /* SGI */


