/* RTcmix  - Copyright (C) 2001  The RTcmix Development Team
   See ``AUTHORS'' for a list of contributors. See ``LICENSE'' for
   the license to this software and for a DISCLAIMER OF ALL WARRANTIES.
*/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>  /* for strerror */
#include <math.h>    /* for fabs */
#include <errno.h>
#include <globals.h>
#include <prototypes.h>
#include <sndlibsupport.h>

#define NORM_FLOAT_FACTOR 0.000030517578


/* --------------------------------------------------------- rtwritesamps --- */
/* Write the current output buffers to an interleaved sound file.
   Currently this supports 16-bit signed integer and 32-bit floating-point
   files, in both byte orders.  Floats have a scaling option that
   forces the normal range of values to fall between -1.0 and 1.0.
*/
int
rtwritesamps()
{
   int            i, n, nsamps, nframes, bytes_per_samp;
   long           nbytes;
   ssize_t        bytes_written;
   unsigned char  *bufp;
   static short   *sbuf = NULL;
   static float   *fbuf = NULL;

   nframes = RTBUFSAMPS;
   nsamps = RTBUFSAMPS * NCHANS;
   bytes_per_samp = mus_data_format_to_bytes_per_sample(output_data_format);

   if (is_float_format) {
      /* Will not have gone through limiter, so do peak check here. */
      if (check_peaks) {
         for (n = 0; n < NCHANS; n++) {
            BufPtr bp = out_buffer[n];
            for (i = 0; i < nframes; i++) {
               double fabsamp = fabs((double) bp[i]);
               if (fabsamp > (double) peaks[n]) {
                  peaks[n] = (BUFTYPE) fabsamp;
                  peaklocs[n] = bufStartSamp + i;         /* frame count */
               }
            }
         }
      }
      if (fbuf == NULL) {        /* first call, so allocate it */
         fbuf = (float *) malloc(nsamps * bytes_per_samp);
         if (fbuf == NULL) {
            fprintf(stderr, "rtwritesamps: can't allocate float buffer\n");
            exit(1);
         }
      }
      bufp = (unsigned char *) fbuf;
   }
   else {
      /* NB: peak check already done in limiter. */
      if (sbuf == NULL) {        /* first call, so allocate it */
         sbuf = (short *) malloc(nsamps * bytes_per_samp);
         if (sbuf == NULL) {
            fprintf(stderr, "rtwritesamps: can't allocate short buffer\n");
            exit(1);
         }
      }
      bufp = (unsigned char *) sbuf;
   }

   switch (output_data_format) {
      case MUS_BSHORT:    
         for (i = 0; i < nframes; i++)
            for (n = 0; n < NCHANS; n++, bufp += bytes_per_samp)
               m_set_big_endian_short(bufp, out_buffer[n][i]);
         bufp = (unsigned char *) sbuf;      /* reset for write call */
         break;
      case MUS_LSHORT:    
         for (i = 0; i < nframes; i++)
            for (n = 0; n < NCHANS; n++, bufp += bytes_per_samp)
               m_set_little_endian_short(bufp, out_buffer[n][i]);
         bufp = (unsigned char *) sbuf;
         break;
      case MUS_BFLOAT:    
         if (normalize_output_floats) {
            for (i = 0; i < nframes; i++)
               for (n = 0; n < NCHANS; n++, bufp += bytes_per_samp)
                  m_set_big_endian_float(bufp,
                                         out_buffer[n][i] * NORM_FLOAT_FACTOR);
         }
         else {
            for (i = 0; i < nframes; i++)
               for (n = 0; n < NCHANS; n++, bufp += bytes_per_samp)
                  m_set_big_endian_float(bufp, out_buffer[n][i]);
         }
         bufp = (unsigned char *) fbuf;
         break;
      case MUS_LFLOAT:    
         if (normalize_output_floats) {
            for (i = 0; i < nframes; i++)
               for (n = 0; n < NCHANS; n++, bufp += bytes_per_samp)
                  m_set_little_endian_float(bufp,
                                         out_buffer[n][i] * NORM_FLOAT_FACTOR);
         }
         else {
            for (i = 0; i < nframes; i++)
               for (n = 0; n < NCHANS; n++, bufp += bytes_per_samp)
                  m_set_little_endian_float(bufp, out_buffer[n][i]);
         }
         bufp = (unsigned char *) fbuf;
         break;
      default:
         fprintf(stderr, "rtwritesamps: unknown output data format!\n");
         exit(1);
         break;
   }

   nbytes = nsamps * bytes_per_samp;

   /* This catches our new case where rtoutput() failed but was ignored */
   if (rtfileit < 0) {
      fprintf(stderr, "rtwritesamps: No output file open (rtoutput failed).\n");
      exit(1);
   }
   
   bytes_written = write(rtoutfile, bufp, nbytes);
   if (bytes_written == -1) {
      fprintf(stderr, "rtwritesamps: %s\n", strerror(errno));
      exit(1);
   }
   /* FIXME: Instead, should probably continue writing... */
   if (bytes_written != nbytes)
      fprintf(stderr, "rtwritesamps: buffer not completely written.");

   return 0;
}


/* ----------------------------------------------------------- rtcloseout --- */
/* Close the output sound file. */
int
rtcloseout()
{
   int result, samps_written;

   if (rtfileit != 1)          /* nothing to close */
      return 0;

   if (check_peaks) {
      result = sndlib_put_header_comment(rtoutfile, peaks, peaklocs, NULL);
      /* any errors reported in this function, but let's keep going */
   }

   /* NOTE: bufStartSamp was incremented again after writing last buffer. */
   samps_written = bufStartSamp * NCHANS;

   result = sndlib_close(rtoutfile, 1, output_header_type,
                                        output_data_format, samps_written);
   if (result == -1) {
      fprintf(stderr, "rtcloseout: Error closing \"%s\" (%s)\n",
                                             rtoutsfname, strerror(errno));
      return -1;
   }

   return 0;
}


